// K-3D
// Copyright (c) 1995-2004, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Remeshing tool based on Harmonic Functions for Quadrilateral Remeshing of Arbitrary Manifolds, S. Dong, S. Kircher, M. Garland
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dsdk/basic_math.h>
#include <k3dsdk/object.h>
#include <k3dsdk/persistence.h>
#include <k3dsdk/property.h>
#include <k3dsdk/measurement.h>
#include <k3dsdk/mesh_filter.h>
#include <k3dsdk/module.h>
#include <k3dsdk/plugins.h>
#include <k3dsdk/utility.h>

namespace libk3dmesh
{

namespace detail
{

extern "C"
{
#include "superlu/src/dsp_defs.h"
}

double cotangent(const double cosine)
{
	const double sine = std::sqrt(1 - cosine*cosine);
	if(sine == 0)
		{
			std::cerr << debug << "Null triangle area" << std::endl;
			return 0;
		}

	return cosine / sine;
}

void solve_2unknown(const double a1, const double b1, const double c1, const double a2, const double b2, const double c2, double& y, double& z)
{
	if(a1 == 0)
		{
			if(b1 != 0)
				{
					z = c1 / b1;
					if(a2 != 0)
						{
							y = (c2 - b2*z) / a2;
							return;
						}
				}
		}
	else
		{
			if(b1 == 0)
				{
					y = c1 / a1;
					if(b2 != 0)
						{
							z = (c2 - a2*y) / b2;
							return;
						}
				}
			else
				{
					if(a2 == 0)
						{
							if(b2 != 0)
								{
									z = c2 / b2;
									y = (c1 - b1*z) / a1;
									return;
								}
						}
					else
						{
							if(b2 == 0)
								{
									y = c2 / a2;
									z = (c1 - a1*y) / b1;
									return;
								}
							else
								{
									// Eliminate y
									z = (a2/a1*c1 - c2) / (a2/a1*b1 - b2);
									y = (c1 - b1*z) / a1;
									return;
								}
						}
				}
		}

	std::cerr << debug << "can't solve system" << std::endl;
	assert_not_reached();

	y = 0;
	z = 0;
}

bool segment_intersection(const double s1x1, const double s1y1, const double s1x2, const double s1y2,
	const double s2x1, const double s2y1, const double s2x2, const double s2y2,
	double& r, double &s)
{
	double denominator = (s1x2 - s1x1)*(s2y2 - s2y1) - (s1y2 - s1y1)*(s2x2 - s2x1);
	if(denominator == 0)
		// Parallel
		return false;

	r = ((s1y1 - s2y1)*(s2x2 - s2x1) - (s1x1 - s2x1)*(s2y2 - s2y1)) / denominator;
	s = ((s1y1 - s2y1)*(s1x2 - s1x1) - (s1x1 - s2x1)*(s1y2 - s1y1)) / denominator;

	return true;
}

bool segment_intersection(const k3d::vector3& A, const k3d::vector3& B, const k3d::vector3& C, const k3d::vector3& D, double& r, double& s)
{
	// Remove X, Y or Z for 2D projection
	double d1x = fabs(B[0] - A[0]);
	double d1y = fabs(B[1] - A[1]);
	double d1z = fabs(B[2] - A[2]);
	double d2x = fabs(D[0] - C[0]);
	double d2y = fabs(D[1] - C[1]);
	double d2z = fabs(D[2] - C[2]);

	unsigned long index1 = 0;
	double min1 = d1x;
	if(d1x < d1y)
		{
			if(d1x < d1z)
				{
					index1 = 0;
					min1 = d1x;
				}
			else
				{
					index1 = 2;
					min1 = d1z;
				}
		}
	else
		{
			if(d1y < d1z)
				{
					index1 = 1;
					min1 = d1y;
				}
			else
				{
					index1 = 2;
					min1 = d1z;
				}
		}

	unsigned long index2 = 0;
	double min2 = d2x;
	if(d2x < d2y)
		{
			if(d2x < d2z)
				{
					index2 = 0;
					min2 = d2x;
				}
			else
				{
					index2 = 2;
					min2 = d2z;
				}
		}
	else
		{
			if(d2y < d2z)
				{
					index2 = 1;
					min2 = d2y;
				}
			else
				{
					index2 = 2;
					min2 = d2z;
				}
		}

	unsigned long index = index1;
	if(min2 < min1)
		index = index2;

	if(index == 0)
		return segment_intersection(A[1], A[2], B[1], B[2], C[1], C[2], D[1], D[2], r, s);
	else if(index == 1)
		return segment_intersection(A[0], A[2], B[0], B[2], C[0], C[2], D[0], D[2], r, s);
	else
		return segment_intersection(A[0], A[1], B[0], B[1], C[0], C[1], D[0], D[1], r, s);

	return true;
}

}

/////////////////////////////////////////////////////////////////////////////
// quadrilateral_remeshing_implementation

class quadrilateral_remeshing_implementation :
	public k3d::mesh_filter<k3d::persistent<k3d::object> >
{
	typedef k3d::mesh_filter<k3d::persistent<k3d::object> > base;

public:
	quadrilateral_remeshing_implementation(k3d::idocument& Document) :
		base(Document),
		m_longitudinal_divisions(k3d::init_name("longitudinal_divisions") + k3d::init_description("Longitudinal divisions [integer]") + k3d::init_value(10) + k3d::init_constraint(k3d::constraint::minimum(1UL)) + k3d::init_precision(0) + k3d::init_step_increment(1) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document)),
		m_latitudinal_divisions(k3d::init_name("latitudinal_divisions") + k3d::init_description("Latitudinal divisions [integer]") + k3d::init_value(10) + k3d::init_constraint(k3d::constraint::minimum(1UL)) + k3d::init_precision(0) + k3d::init_step_increment(1) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document)),
		m_min_index(k3d::init_name("min_index") + k3d::init_description("Min index [integer]") + k3d::init_value(513) + k3d::init_constraint(k3d::constraint::minimum(0UL)) + k3d::init_precision(0) + k3d::init_step_increment(1) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document)),
		m_max_index(k3d::init_name("max_index") + k3d::init_description("Max index [integer]") + k3d::init_value(214) + k3d::init_constraint(k3d::constraint::minimum(0UL)) + k3d::init_precision(0) + k3d::init_step_increment(1) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document)),
		m_h1(k3d::init_name("spacing1") + k3d::init_description("Spacing 1 [number]") + k3d::init_value(0.5) + k3d::init_precision(2) + k3d::init_step_increment(0.1) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document)),
		m_h2(k3d::init_name("spacing2") + k3d::init_description("Spacing2 [number]") + k3d::init_value(k3d::radians<double>(0.0)) + k3d::init_precision(2) + k3d::init_step_increment(0.01) + k3d::init_units(typeid(k3d::measurement::angle)) + k3d::init_document(Document)),
		m_curvature_sensibility(k3d::init_name("curvature_sensibility") + k3d::init_description("Curvature sensibility [number]") + k3d::init_value(1.0) + k3d::init_precision(2) + k3d::init_step_increment(0.01) + k3d::init_units(typeid(k3d::measurement::scalar)) + k3d::init_document(Document))
	{
		enable_serialization(k3d::persistence::proxy(m_longitudinal_divisions));
		enable_serialization(k3d::persistence::proxy(m_latitudinal_divisions));
		enable_serialization(k3d::persistence::proxy(m_min_index));
		enable_serialization(k3d::persistence::proxy(m_max_index));
		enable_serialization(k3d::persistence::proxy(m_h1));
		enable_serialization(k3d::persistence::proxy(m_h2));
		enable_serialization(k3d::persistence::proxy(m_curvature_sensibility));

		register_property(m_longitudinal_divisions);
		register_property(m_latitudinal_divisions);
		register_property(m_min_index);
		register_property(m_max_index);
		register_property(m_h1);
		register_property(m_h2);
		register_property(m_curvature_sensibility);

		m_input_mesh.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_input_change));
		m_longitudinal_divisions.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_reset_geometry));
		m_latitudinal_divisions.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_reset_geometry));
		m_min_index.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_input_change));
		m_max_index.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_input_change));
		m_h1.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_reset_geometry));
		m_h2.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_reset_geometry));
		m_curvature_sensibility.changed_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_reset_geometry));
		m_output_mesh.need_data_signal().connect(SigC::slot(*this, &quadrilateral_remeshing_implementation::on_create_geometry));

		m_input_changed = false;
	}

	typedef unsigned long index_t;

	void on_reset_geometry()
	{
		m_output_mesh.reset();
	}

	void on_input_change()
	{
		m_input_changed = true;
	}

	// Mesh preprocessing : computed only once
	void preprocess_geometry(const k3d::mesh& input)
	{
		// Create a mapping of points-to-zero-based-indices; store points ...
		std::map<k3d::point*, unsigned long> point_map;
		unsigned long index = 0;
		m_points.clear();
		m_points.reserve(input.points.size());
		for(k3d::mesh::points_t::const_iterator point = input.points.begin(); point != input.points.end(); ++point)
			{
				const k3d::vector3 p = (*point)->position;
				m_points.push_back(p);
//std::cerr << "P " << index << " : " << p << std::endl;
				point_map[*point] = index;
				++index;
			}

		const unsigned long points_size = m_points.size();

		// Construct piecewise-linear scalar field over vertices
		// TODO: make sure it is a (closed) manifold, triangulated and no extra point

		typedef std::vector<index_t> polygon_t;

		typedef std::set<index_t> index_t_set_t;
		typedef std::vector<index_t_set_t> adjacency_t;
		adjacency_t adjacency_sets;
		adjacency_sets.resize(m_points.size());
		std::vector< std::map<index_t, double> > cotangents;
		cotangents.resize(m_points.size());

		typedef std::pair<unsigned long, unsigned long> edge_t;
		std::map<edge_t, unsigned long> edges_to_triangles;

		points_t triangle_normals;

		unsigned long triangle_number = 0;
		m_triangles.clear();

		for(k3d::mesh::polyhedra_t::const_iterator polyhedron = input.polyhedra.begin(); polyhedron != input.polyhedra.end(); ++polyhedron)
			for(k3d::polyhedron::faces_t::const_iterator face = (*polyhedron)->faces.begin(); face != (*polyhedron)->faces.end(); ++face)
			{
				polygon_t triangle;
				points_t coords;

				k3d::split_edge* first = (*face)->first_edge;
				if(!first)
					// Skip empty faces
					continue;

				triangle.push_back(point_map[first->vertex]);
				coords.push_back(first->vertex->position);

				k3d::split_edge* current_edge = first->face_clockwise;
				while(current_edge)
				{
					triangle.push_back(point_map[current_edge->vertex]);
					coords.push_back(current_edge->vertex->position);

					current_edge = current_edge->face_clockwise;
					if(current_edge == first)
						break;
				}

				// Sanity check
				assert_warning(triangle.size() == 3);

				// Compute discrete harmonic weights (for edges)
				coords[0].Normalize();
				coords[1].Normalize();
				coords[2].Normalize();
				k3d::vector3 v21 = coords[1] - coords[2];
				k3d::vector3 v20 = coords[0] - coords[2];
				v21.Normalize();
				v20.Normalize();
				double cot12 = detail::cotangent(v21 * v20);

				k3d::vector3 v02 = coords[2] - coords[0];
				k3d::vector3 v01 = coords[1] - coords[0];
				v02.Normalize();
				v01.Normalize();
				double cot23 = detail::cotangent(v02 * v01);

				k3d::vector3 v10 = coords[0] - coords[1];
				k3d::vector3 v12 = coords[2] - coords[1];
				v10.Normalize();
				v12.Normalize();
				double cot31 = detail::cotangent(v10 * v12);

				// Build adjacency array
				index_t v1 = triangle[0];
				index_t v2 = triangle[1];
				index_t v3 = triangle[2];

				adjacency_sets[v1].insert(v2);
				adjacency_sets[v2].insert(v1);
				cotangents[v1].insert(std::pair<index_t, double>(v2, cot12));

				adjacency_sets[v2].insert(v3);
				adjacency_sets[v3].insert(v2);
				cotangents[v2].insert(std::pair<index_t, double>(v3, cot23));

				adjacency_sets[v3].insert(v1);
				adjacency_sets[v1].insert(v3);
				cotangents[v3].insert(std::pair<index_t, double>(v1, cot31));

				// Build triangles list
				triangle_t real_triangle;
				real_triangle[0] = v1;
				real_triangle[1] = v2;
				real_triangle[2] = v3;
				m_triangles.push_back(real_triangle);

				// Add triangle normal
				k3d::vector3 normal = k3d::normal((*face)->first_edge);
				normal.Normalize();
				triangle_normals.push_back(normal);

				// Build edges -> triangles map
				edges_to_triangles.insert(std::pair<edge_t, unsigned long>(edge_t(v1, v2), triangle_number));
				edges_to_triangles.insert(std::pair<edge_t, unsigned long>(edge_t(v2, v3), triangle_number));
				edges_to_triangles.insert(std::pair<edge_t, unsigned long>(edge_t(v3, v1), triangle_number));

				++triangle_number;
			}

		unsigned long vn = 0;

		std::vector<double> non_zero_coefficients;
		std::vector<index_t> row_indices;
		std::vector<index_t> coefficient_row_indices;
		coefficient_row_indices.push_back(0);

		m_flow_min = m_min_index.property_value();
		m_flow_max = m_max_index.property_value();
		double min_value = -1000;
		double max_value = 1000;

		for(adjacency_t::const_iterator va = adjacency_sets.begin(); va != adjacency_sets.end(); ++va)
			{
				unsigned long row_values = row_indices.size();

				if(m_flow_min == vn || m_flow_max == vn)
				{
					// Aij = deltaij if i in C
					non_zero_coefficients.push_back(1);
					row_indices.push_back(vn);

					row_values += 1;
				}
				else
				{
					polygon_t row;
					for(index_t_set_t::const_iterator sa = va->begin(); sa != va->end(); ++sa)
						row.push_back(*sa);
					assert_warning(row.size());
					std::sort(row.begin(), row.end());

					polygon_t::const_iterator row_item = row.begin();
					polygon_t::const_iterator row_end = row.end();
					bool passed_diagonal = false;
					double w_sum = 0;
					unsigned long diagonal_pos = 0;
					for(; row_item != row_end; ++row_item)
						{
							index_t adjacent_vertex = *row_item;

							// Lij = 1 if i == j
							if(adjacent_vertex > vn && !passed_diagonal)
								{
									diagonal_pos = non_zero_coefficients.size();
									non_zero_coefficients.push_back(1);
									row_indices.push_back(vn);

									passed_diagonal = true;
								}

							// Opposite angle cotangent
							double cot1 = cotangents[vn][adjacent_vertex];
							// Opposite angle cotangent in other triangle
							double cot2 = cotangents[adjacent_vertex][vn];
							double w = cot1 + cot2;
							w_sum += w;

							non_zero_coefficients.push_back(w);
							row_indices.push_back(adjacent_vertex);
						}

					// Lij = 1 if i == j
					w_sum = -w_sum;
					if(!passed_diagonal)
						{
							non_zero_coefficients.push_back(w_sum);
							row_indices.push_back(vn);
						}
					else
						{
							non_zero_coefficients[diagonal_pos] = w_sum;
						}

					row_values += row.size() + 1;
				}

				coefficient_row_indices.push_back(row_values);

				++vn;
			}

		// Create SuperLU matrices
		detail::SuperMatrix A, L, U, B;
		double   *a, *rhs;
		int      *asub, *xa;
		int      *perm_r; // row permutations from partial pivoting
		int      *perm_c; // column permutation vector
		int      nrhs, info, i, m, n, nnz;
		detail::superlu_options_t options;
		detail::SuperLUStat_t stat;

		// Initialize matrix A
		m = n = points_size;
		nnz = non_zero_coefficients.size();
		return_if_fail(a = detail::doubleMalloc(nnz));
		return_if_fail(asub = detail::intMalloc(nnz));
		return_if_fail(xa = detail::intMalloc(n+1));

		assert_warning(coefficient_row_indices.size() == points_size + 1);


		std::vector<double>::const_iterator nz = non_zero_coefficients.begin();
		std::vector<double>::const_iterator nz_end = non_zero_coefficients.end();
		double* a_p = a;
		for(; nz != nz_end; ++nz)
			*a_p++ = *nz;

		std::vector<index_t>::const_iterator ri = row_indices.begin();
		std::vector<index_t>::const_iterator ri_end = row_indices.end();
		int* asub_p = asub;
		for(; ri != ri_end; ++ri)
			*asub_p++ = *ri;

		std::vector<index_t>::const_iterator cr = coefficient_row_indices.begin();
		std::vector<index_t>::const_iterator cr_end = coefficient_row_indices.end();
		int* xa_p = xa;
		for(; cr != cr_end; ++cr)
			*xa_p++ = *cr;

		// Create matrix A in the format expected by SuperLU
		detail::dCreate_CompCol_Matrix(&A, m, n, nnz, a, asub, xa, detail::SLU_NR, detail::SLU_D, detail::SLU_GE);

		// Create right-hand side matrix B
		nrhs = 1;
		return_if_fail(rhs = detail::doubleMalloc(m * nrhs));
		for (i = 0; i < m; ++i) rhs[i] = 0.0;
		rhs[m_flow_min] = min_value;
		rhs[m_flow_max] = max_value;
		detail::dCreate_Dense_Matrix(&B, m, nrhs, rhs, m, detail::SLU_DN, detail::SLU_D, detail::SLU_GE);

		return_if_fail(perm_r = detail::intMalloc(m));
		return_if_fail(perm_c = detail::intMalloc(n));

		// Set the default input options
		detail::set_default_options(&options);
		options.ColPerm = detail::NATURAL;

		// Initialize the statistics variables
		detail::StatInit(&stat);

		detail::dgssv(&options, &A, perm_c, perm_r, &L, &U, &B, &stat, &info);

		// Print solution
//std::cerr << "Scalar field" << std::endl;
//		for (i = 0; i < m; ++i) std::cerr << i << " : " << rhs[i] << std::endl;

		// Save values, find near zero values
		m_scalar_field.clear();
		m_scalar_field.reserve(m);
		unsigned long min_positive_index = 0;
		double min_positive_value = max_value;
		unsigned long max_negative_index = 0;
		double max_negative_value = min_value;
		for(i = 0; i < m; ++i)
			{
				double value = rhs[i];
				m_scalar_field.push_back(value);

				if(value >= 0)
					if(value < min_positive_value)
						{
							min_positive_index = i;
							min_positive_value = value;
						}
				else
					if(value > max_negative_value)
						{
							max_negative_index = i;
							max_negative_value = value;
						}
			}

		// De-allocate storage
		detail::SUPERLU_FREE (rhs);
		detail::SUPERLU_FREE (perm_r);
		detail::SUPERLU_FREE (perm_c);
		detail::Destroy_CompCol_Matrix(&A);
		detail::Destroy_SuperMatrix_Store(&B);
		detail::Destroy_SuperNode_Matrix(&L);
		detail::Destroy_CompCol_Matrix(&U);
		detail::StatFree(&stat);

		// List triangle neighbours, compute orthogonal and tangent fields (g1 and g2)
		m_triangle_neighbours.clear();
		m_orthogonal_vectors.clear();
		m_tangent_vectors.clear();
		triangle_list_t::const_iterator triangle_i = m_triangles.begin();
		triangle_list_t::const_iterator triangles_end = m_triangles.end();
		triangle_number = 0;
		unsigned long middle_triangle = 0;
		for(; triangle_i != triangles_end; ++triangle_i)
			{
				index_t v1 = (*triangle_i)[0];
				index_t v2 = (*triangle_i)[1];
				index_t v3 = (*triangle_i)[2];
//std::cerr << "Face " << triangle_number << " : " << v1 << ", " << v2 << ", " << v3 << std::endl;

				if(v1 == min_positive_index || v2 == min_positive_index || v3 == min_positive_index)
					middle_triangle = triangle_number;

				// Store neighbours
				triangle_t neighbours;
				neighbours[0] = edges_to_triangles[edge_t(v2, v1)];
				neighbours[1] = edges_to_triangles[edge_t(v3, v2)];
				neighbours[2] = edges_to_triangles[edge_t(v1, v3)];
//std::cerr << "Neighbours " << " : " << neighbours[0] << ", " << neighbours[1] << ", " << neighbours[2] << std::endl;

				m_triangle_neighbours.push_back(neighbours);

				// Gradient vector g1
				k3d::vector3 xi = m_points[v1];
				k3d::vector3 xj = m_points[v2];
				k3d::vector3 xk = m_points[v3];
/*std::cerr << "p1 : " << xi << std::endl;
std::cerr << "p2 : " << xj << std::endl;
std::cerr << "p3 : " << xk << std::endl;*/
				k3d::vector3 normal = triangle_normals[triangle_number];

				double ui = m_scalar_field[v1];
				double uj = m_scalar_field[v2];
				double uk = m_scalar_field[v3];

				double xji = xj[0] - xi[0];
				double yji = xj[1] - xi[1];
				double zji = xj[2] - xi[2];
				double xkj = xk[0] - xj[0];
				double ykj = xk[1] - xj[1];
				double zkj = xk[2] - xj[2];

				double ni = normal[0];
				double nj = normal[1];
				double nk = normal[2];
				double uji = uj - ui;
				double ukj = uk - uj;

				double x = 0;
				double y = 0;
				double z = 0;
				if(ni != 0)
					{
						// Eliminate x
						double nyji = yji;
						double nzji = zji;
						double nuji = uji;
						if(xji != 0)
							{
								double coef = ni/xji;
								nyji = coef*yji - nj;
								nzji = coef*zji - nk;
								nuji *= coef;
							}

						double nykj = ykj;
						double nzkj = zkj;
						double nukj = ukj;
						if(xkj != 0)
							{
								double coef = ni/xkj;
								nykj = coef*ykj - nj;
								nzkj = coef*zkj - nk;
								nukj *= coef;
							}

						detail::solve_2unknown(nyji, nzji, nuji, nykj, nzkj, nukj, y, z);
						x = - (nj*y + nk*z) / ni;
					}
				else if(nj != 0)
					{
						// Eliminate y
						double nxji = xji;
						double nzji = zji;
						double nuji = uji;
						if(yji != 0)
							{
								double coef = nj/yji;
								nxji = coef*xji - ni;
								nzji = coef*zji - nk;
								nuji *= coef;
							}

						double nxkj = xkj;
						double nzkj = zkj;
						double nukj = ukj;
						if(ykj != 0)
							{
								double coef = nj/ykj;
								nxkj = coef*xkj - ni;
								nzkj = coef*zkj - nk;
								nukj *= coef;
							}

						detail::solve_2unknown(nxji, nzji, nuji, nxkj, nzkj, nukj, x, z);
						y = - (ni*x + nk*z) / nj;
					}
				else if(nk != 0)
					{
						// Eliminate z
						double nxji = xji;
						double nyji = yji;
						double nuji = uji;
						if(zji != 0)
							{
								double coef = nk/zji;
								nxji = coef*xji - ni;
								nyji = coef*yji - nj;
								nuji *= coef;
							}

						double nxkj = xkj;
						double nykj = ykj;
						double nukj = ukj;
						if(zkj != 0)
							{
								double coef = nk/zkj;
								nxkj = coef*xkj - ni;
								nykj = coef*ykj - nj;
								nukj *= coef;
							}

						detail::solve_2unknown(nxji, nyji, nuji, nxkj, nykj, nukj, x, y);
						z = - (ni*x + nj*z) / nk;
					}
				else
					{
						std::cerr << debug << "Null normal" << std::endl;
						assert_not_reached();
					}

				k3d::vector3 g1(x, y, z);
g1.Normalize();
				m_orthogonal_vectors.push_back(g1);

				// Orthogonal vector g2
				k3d::vector3 g2 = normal ^ g1;
g2.Normalize();
				m_tangent_vectors.push_back(g2);

const k3d::vector3 bary = (xi + xj + xk) / 3;
//std::cerr << "bary : " << bary << std::endl;
norms1.push_back(bary);
norms1.push_back(bary + .2*g1);
norms2.push_back(bary);
norms2.push_back(bary + g2*0.2);
//std::cerr << "g1 " << " : " << g1 << std::endl;
//std::cerr << "g2 " << " : " << g2 << std::endl;

				// Update triangle number
				++triangle_number;
			}

		m_middle_triangle = middle_triangle;

		on_reset_geometry();
	}

	k3d::mesh* on_create_geometry()
	{
		// Get the input geometry
		k3d::mesh* const input = m_input_mesh.property_value();
		if(!input)
			return 0;

		// Check for new mesh
		if(m_input_changed)
			{
				preprocess_geometry(*input);
				m_input_changed = false;
			}

		// Sanity checks
		return_val_if_fail(m_scalar_field.size() == m_points.size(), 0);

		// Create output geometry ...
		k3d::mesh* const output = new k3d::mesh();

		// Clear any previous work
		m_gradient_flows.clear();

		// Create first gradient flow
		m_test_flow_distance = false;
		index_t middle_triangle = m_middle_triangle;
		m_current_triangle = middle_triangle;

		// Start from our "middle triangle"'s barycenter
		unsigned long v1 = m_triangles[middle_triangle][0];
		unsigned long v2 = m_triangles[middle_triangle][1];
		unsigned long v3 = m_triangles[middle_triangle][2];
		k3d::vector3 barycenter = (m_points[v1] + m_points[v2] + m_points[v3]) / 3;
		double bary_u = (m_scalar_field[v1] + m_scalar_field[v2] + m_scalar_field[v3]) / 3;

		// Init
		m_gradient_flow_points.clear();
		m_gradient_flow_points.resize(m_triangles.size());

		// Get and save gradient flow
		points_t first_flow;
		indices_t first_triangles;
		std::vector<double> first_values;
		gradient_flow_line(middle_triangle, barycenter, bary_u, first_flow, first_triangles, first_values);
		m_gradient_flows.push_back(first_flow);
		m_gradient_flows_triangles.push_back(first_triangles);
		m_gradient_flows_values.push_back(first_values);

		// Trace other gradient flow lines using h1 as flow distance
		const double h1 = m_h1.property_value();
		m_gradient_distance =  h1 * 0.5;

		// Go through every flow we create
		unsigned long current_flow = 0;
		while(current_flow < m_gradient_flows.size())
			{
				std::vector<k3d::vector3> seed_points;
				std::vector<index_t> seed_triangles;
				std::vector<double> seed_values;

				// Find a seed point from either side for each flow point
				for(unsigned long n = 1; n < m_gradient_flows[current_flow].size() - 1; ++n)
					{
						m_intersection = m_gradient_flows[current_flow][n];
						m_current_triangle = m_gradient_flows_triangles[current_flow][n];
						m_new_u = m_gradient_flows_values[current_flow][n];
						k3d::vector3 new_point;
						index_t new_triangle;
						if(isoparametric_seed_point(h1, new_point, new_triangle, true))
							{
								seed_points.push_back(new_point);
								seed_triangles.push_back(new_triangle);
								seed_values.push_back(m_new_u);
							}

						m_intersection = m_gradient_flows[current_flow][n];
						m_current_triangle = m_gradient_flows_triangles[current_flow][n];
						m_new_u = m_gradient_flows_values[current_flow][n];
						if(isoparametric_seed_point(h1, new_point, new_triangle, false))
							{
								seed_points.push_back(new_point);
								seed_triangles.push_back(new_triangle);
								seed_values.push_back(m_new_u);
							}
					}

				// Draw new flows
				m_test_flow_distance = true;
				for(unsigned long n = 0; n < seed_points.size(); ++n)
					{
						m_intersection = seed_points[n];
						m_current_triangle = seed_triangles[n];
						m_new_u = seed_values[n];
						if(flow_distance_ok())
							{
								points_t new_flow;
								indices_t new_triangles;
								std::vector<double> new_values;
								gradient_flow_line(seed_triangles[n], seed_points[n], seed_values[n], new_flow, new_triangles, new_values);
								if(new_flow.size())
									{
										m_gradient_flows.push_back(new_flow);
										m_gradient_flows_triangles.push_back(new_triangles);
										m_gradient_flows_values.push_back(new_values);
									}
							}
					}

				++current_flow;
			}


		// Create first isoparametric flow
		m_current_triangle = middle_triangle;
		m_intersection = barycenter;
		m_new_u = bary_u;

		points_t iso_flow;
		indices_t iso_indices;
		isoparametric_flow_line(iso_flow, iso_indices);

/*
		// Create other gradient flows
		points_t salient_points;
		indices_t salient_indices;
		const unsigned long divisions = m_latitudinal_divisions.property_value();
		flow_divisions(iso_flow, divisions, salient_points, salient_indices);

		for(unsigned long n = 0; n < divisions; ++n)
			{
				unsigned long triangle = iso_indices[salient_indices[n]];
				k3d::vector3 point = salient_points[n];

				// New flow
				points_t new_flow;
				indices_t new_indices;
				std::vector<double> new_values;
				gradient_flow_line(triangle, point, bary_u, new_flow, new_indices, new_values);

				// New group
				k3d::linear_curve_group* const group = new k3d::linear_curve_group();
				k3d::linear_curve* const curve = new k3d::linear_curve();
				for(unsigned long i = 0; i < new_flow.size(); ++i)
					{
						const k3d::vector3 p = new_flow[i];
						k3d::point* const p_p = new k3d::point(p);
						output->points.push_back(p_p);

						curve->control_points.push_back(p_p);
					}

				group->curves.push_back(curve);
				output->linear_curve_groups.push_back(group);
			}

		// Create other isoparametric flows
		points_t iso_salient_points;
		indices_t iso_salient_indices;
		const unsigned long iso_divisions = m_longitudinal_divisions.property_value();
		flow_divisions(first_flow, iso_divisions, iso_salient_points, iso_salient_indices);

		//for(unsigned long n = 0; n < iso_divisions; ++n)
		for(unsigned long n = 0; n < 3; ++n)
			{
				m_current_triangle = first_triangles[iso_salient_indices[n]];
				m_intersection = iso_salient_points[n];
				m_new_u = first_values[iso_salient_indices[n]];

				// New flow
				points_t new_flow;
				indices_t new_indices;
				isoparametric_flow_line(new_flow, new_indices);

				// New group
				k3d::linear_curve_group* const group = new k3d::linear_curve_group();
				k3d::linear_curve* const curve = new k3d::linear_curve();
				for(unsigned long i = 0; i < new_flow.size(); ++i)
					{
						const k3d::vector3 p = new_flow[i];
						k3d::point* const p_p = new k3d::point(p);
						output->points.push_back(p_p);

						curve->control_points.push_back(p_p);
					}

				group->curves.push_back(curve);
				output->linear_curve_groups.push_back(group);
			}
*/

		// TODO: form polygons







		// Draw gradient flows
		for(unsigned long g = 0; g < m_gradient_flows.size(); ++g)
			{
				k3d::linear_curve_group* const group = new k3d::linear_curve_group();
				k3d::linear_curve* const curve = new k3d::linear_curve();
				for(unsigned long i = 0; i < m_gradient_flows[g].size(); ++i)
					{
						const k3d::vector3 p = m_gradient_flows[g][i];
						k3d::point* const p_p = new k3d::point(p);
						output->points.push_back(p_p);

						curve->control_points.push_back(p_p);
					}

				group->curves.push_back(curve);
				output->linear_curve_groups.push_back(group);
			}

		// Iso-flow
		k3d::linear_curve_group* const group3 = new k3d::linear_curve_group();
		k3d::linear_curve* const curve3 = new k3d::linear_curve();
		iso_flow.push_back(iso_flow[0]);
		for(unsigned long i = 0; i < iso_flow.size(); ++i)
			{
				const k3d::vector3 p = iso_flow[i];
				k3d::point* const p_p = new k3d::point(p);
				output->points.push_back(p_p);

				curve3->control_points.push_back(p_p);
			}

		group3->curves.push_back(curve3);
		output->linear_curve_groups.push_back(group3);


/*
		// NORMS2
		k3d::linear_curve_group* const group2 = new k3d::linear_curve_group();
		for(unsigned long i = 0; i < norms2.size() / 2; ++i)
			{
				k3d::linear_curve* const curve = new k3d::linear_curve();

				const k3d::vector3 base = norms2[i*2];
				const k3d::vector3 tip = norms2[i*2 + 1];

				k3d::point* const base_p = new k3d::point(base);
				k3d::point* const tip_p = new k3d::point(tip);

				output->points.push_back(base_p);
				output->points.push_back(tip_p);

				curve->control_points.push_back(base_p);
				curve->control_points.push_back(tip_p);

				group2->curves.push_back(curve);
			}
		output->linear_curve_groups.push_back(group2);
*/

/*
		k3d::linear_curve_group* const groupD = new k3d::linear_curve_group();
		for(unsigned long i = 0; i < m_debug_vector_list.size() / 2; ++i)
			{
				k3d::linear_curve* const curve = new k3d::linear_curve();

				const k3d::vector3 base = m_debug_vector_list[i*2];
				const k3d::vector3 tip = m_debug_vector_list[i*2 + 1];

				k3d::point* const base_p = new k3d::point(base);
				k3d::point* const tip_p = new k3d::point(tip);

				output->points.push_back(base_p);
				output->points.push_back(tip_p);

				curve->control_points.push_back(base_p);
				curve->control_points.push_back(tip_p);

				groupD->curves.push_back(curve);
			}
		output->linear_curve_groups.push_back(groupD);
*/

/*
		// NORMS1
		k3d::linear_curve_group* const group1 = new k3d::linear_curve_group();
		for(unsigned long i = 0; i < norms1.size() / 2; ++i)
			{
				k3d::linear_curve* const curve = new k3d::linear_curve();

				const k3d::vector3 base = norms1[i*2];
				const k3d::vector3 tip = norms1[i*2 + 1];

				k3d::point* const base_p = new k3d::point(base);
				k3d::point* const tip_p = new k3d::point(tip);

				output->points.push_back(base_p);
				output->points.push_back(tip_p);

				curve->control_points.push_back(base_p);
				curve->control_points.push_back(tip_p);

				group1->curves.push_back(curve);
			}
		output->linear_curve_groups.push_back(group1);
*/

/*
		std::map<unsigned long, k3d::point*> point_map;
		for(unsigned int v = 0; v < m->vert_count(); v++)
			{
				k3d::vector3 position(m->vertex(v));
				position[0] = -position[0];

				k3d::point* point = new k3d::point(position);
				output->points.push_back(point);

				point_map[v] = point;
			}

		k3d::polyhedron* polyhedron = new k3d::polyhedron();
		return_val_if_fail(polyhedron, 0);
		output->polyhedra.push_back(polyhedron);
		for(unsigned int f = 0; f < m->face_count(); f++)
			if(m->face_is_valid(f))
				{
					unsigned long i1 = m->face(f).v[0];
					unsigned long i2 = m->face(f).v[1];
					unsigned long i3 = m->face(f).v[2];

					// Create a triangle ...
					k3d::split_edge* edge1 = new k3d::split_edge(point_map[i1]);
					k3d::split_edge* edge2 = new k3d::split_edge(point_map[i2]);
					k3d::split_edge* edge3 = new k3d::split_edge(point_map[i3]);

					// Invert face (because of coordinate system switch)
					edge1->face_clockwise = edge3;
					edge3->face_clockwise = edge2;
					edge2->face_clockwise = edge1;

					polyhedron->edges.push_back(edge1);
					polyhedron->edges.push_back(edge2);
					polyhedron->edges.push_back(edge3);

					k3d::face* const face = new k3d::face(edge1);
					return_val_if_fail(face, 0);

					polyhedron->faces.push_back(face);
				}

		// Copy materials ...
		if(input->polyhedra.size() && output->polyhedra.size())
			output->polyhedra.front()->material = input->polyhedra.front()->material;
*/

		return output;
	}

	k3d::iplugin_factory& factory()
	{
		return get_factory();
	}

	static k3d::iplugin_factory& get_factory()
	{
		static k3d::plugin_factory<
			k3d::document_plugin<quadrilateral_remeshing_implementation>,
				k3d::interface_list<k3d::imesh_source,
				k3d::interface_list<k3d::imesh_sink > > > factory(
				k3d::uuid(0xde6494ab, 0x1c4d448d, 0xa6a2657e, 0x1751fc93),
				"QuadrilateralRemeshing",
				"Quadrilateral Remeshing for Closed Manifolds",
				"Objects",
				k3d::iplugin_factory::EXPERIMENTAL);

		return factory;
	}

private:
	typedef std::vector<k3d::vector3> points_t;
	typedef std::vector<index_t> indices_t;

	typedef enum
	{
		EDGE,
		POINT,
		NONE
	} intersection_t;

	unsigned long test_intersection(unsigned long v1, unsigned long v2, unsigned long v3, unsigned long edge_index, intersection_t& intersection_type)
	{
		k3d::vector3 p1 = m_points[v1];
		k3d::vector3 p2 = m_points[v2];
		double u1 = m_scalar_field[v1];
		double u2 = m_scalar_field[v2];

		double r;
		double s;
		if(detail::segment_intersection(m_intersection, m_intersection + m_direction, p1, p2, r, s))
			{
				m_iedge1 = v1;
				m_iedge2 = v2;

				if(s == 0)
					{
						m_intersection = p1;
						m_intersection_index = v1;
						m_new_u = u1;
						intersection_type = POINT;
						return m_triangle_neighbours[m_current_triangle][edge_index];
					}
				else if(s == 1)
					{
						m_intersection = p2;
						m_intersection_index = v2;
						m_new_u = u2;
						intersection_type = POINT;
						return m_triangle_neighbours[m_current_triangle][(edge_index + 1) % 3];
					}
				else if(r < 0)
					{
						m_missed = false;
					}
				else if(r >= 0 && s > 0 && s < 1)
					{
						m_intersection = p1 + s * (p2 - p1);
						m_new_u = u1 + s*(u2 - u1);
						m_iedge1 = v1;
						m_iedge2 = v2;
						intersection_type = EDGE;
						return m_triangle_neighbours[m_current_triangle][edge_index];
					}
				else if(r >= 0)
					{
						if(s < 0)
							{
								m_missed = true;
								m_missed_index = v1;
							}
						if(s > 1)
							{
								m_missed = true;
								m_missed_index = v2;
							}
					}
			}

		intersection_type = NONE;
		return 0;
	}

	// Turn around v1 to find which triangle is crossed by the flow
	void vertex_case(unsigned long v1, unsigned long v_start, bool up, intersection_t& intersection_type)
	{
		unsigned long v_start2 = v_start;
		unsigned long v2 = v_start;
		bool prev_missed = false;
		unsigned long prev_missed_index = v1;

		do
			{
				index_t p1 = m_triangles[m_current_triangle][0];
				index_t p2 = m_triangles[m_current_triangle][1];
				index_t p3 = m_triangles[m_current_triangle][2];

				intersection_type = NONE;
				if(v1 == p1)
					{
						unsigned long new_triangle = test_intersection(p2, p3, p1, 1, intersection_type);
						if(intersection_type == NONE)
							{
								if(v2 == p2)
									v_start2 = p3;

								v2 = p2;

								if(m_missed)
									{
										if(prev_missed)
											{
												if(m_missed_index == prev_missed_index)
													{
														m_intersection = m_points[m_missed_index];
														m_intersection_index = m_missed_index;
														m_new_u = m_scalar_field[m_missed_index];
														intersection_type = POINT;
														return;
													}
											}

										prev_missed = true;
										prev_missed_index = m_missed_index;
									}

								m_current_triangle = m_triangle_neighbours[m_current_triangle][0];
								m_direction = up ? m_orthogonal_vectors[m_current_triangle] : -m_orthogonal_vectors[m_current_triangle];
							}
						else if(intersection_type == POINT)
							{
								return;
							}
						else if(intersection_type == EDGE)
							{
								m_current_triangle = new_triangle;
								return;
							}
					}
				else if(v1 == p2)
					{
						unsigned long new_triangle = test_intersection(p3, p1, p2, 2, intersection_type);
						if(intersection_type == NONE)
							{
								if(v2 == p3)
									v_start2 = p1;

								v2 = p3;

								if(m_missed)
									{
										if(prev_missed)
											{
												if(m_missed_index == prev_missed_index)
													{
														m_intersection = m_points[m_missed_index];
														m_intersection_index = m_missed_index;
														m_new_u = m_scalar_field[m_missed_index];
														intersection_type = POINT;
														return;
													}
											}

										prev_missed = true;
										prev_missed_index = m_missed_index;
									}

								m_current_triangle = m_triangle_neighbours[m_current_triangle][1];
								m_direction = up ? m_orthogonal_vectors[m_current_triangle] : -m_orthogonal_vectors[m_current_triangle];
							}
						else if(intersection_type == POINT)
							{
								return;
							}
						else if(intersection_type == EDGE)
							{
								m_current_triangle = new_triangle;
								return;
							}
					}
				else if(v1 == p3)
					{
						unsigned long new_triangle = test_intersection(p1, p2, p3, 0, intersection_type);
						if(intersection_type == NONE)
							{
								if(v2 == p1)
									v_start2 = p2;

								v2 = p1;

								if(m_missed)
									{
										if(prev_missed)
											{
												if(m_missed_index == prev_missed_index)
													{
														m_intersection = m_points[m_missed_index];
														m_intersection_index = m_missed_index;
														m_new_u = m_scalar_field[m_missed_index];
														intersection_type = POINT;
														return;
													}
											}

										prev_missed = true;
										prev_missed_index = m_missed_index;
									}

								m_current_triangle = m_triangle_neighbours[m_current_triangle][2];
								m_direction = up ? m_orthogonal_vectors[m_current_triangle] : -m_orthogonal_vectors[m_current_triangle];
							}
						else if(intersection_type == POINT)
							{
								return;
							}
						else if(intersection_type == EDGE)
							{
								m_current_triangle = new_triangle;
								return;
							}
					}
				else
					assert_not_reached();

			}
		while(v2 != v_start2);
	}

	// Regular case
	void edge_intersection(intersection_t& intersection_type)
	{
		index_t v1 = m_triangles[m_current_triangle][0];
		index_t v2 = m_triangles[m_current_triangle][1];
		index_t v3 = m_triangles[m_current_triangle][2];

		unsigned long other_v_pos = 0;
		if(v1 == m_iedge1 || v1 == m_iedge2)
			{
				if(v2 == m_iedge1 || v2 == m_iedge2)
					other_v_pos = 2;
				else
					other_v_pos = 1;
			}

		unsigned long neighbouring_triangle;
		if(other_v_pos == 0)
			{
				neighbouring_triangle = test_intersection(v1, v2, v3, 0, intersection_type);
				if(intersection_type == NONE)
					neighbouring_triangle = test_intersection(v3, v1, v2, 2, intersection_type);
			}
		else if(other_v_pos == 1)
			{
				neighbouring_triangle = test_intersection(v1, v2, v3, 0, intersection_type);
				if(intersection_type == NONE)
					neighbouring_triangle = test_intersection(v2, v3, v1, 1, intersection_type);
			}
		else
			{
				neighbouring_triangle = test_intersection(v2, v3, v1, 1, intersection_type);
				if(intersection_type == NONE)
					neighbouring_triangle = test_intersection(v3, v1, v2, 2, intersection_type);
			}

		// Edge case
		if(intersection_type != NONE)
			{
				m_current_triangle = neighbouring_triangle;
				return;
			}

		if(other_v_pos == 0)
			{
				intersection_type = POINT;
				k3d::vector3 p2 = m_points[v2];
				k3d::vector3 p3 = m_points[v3];
				if((p3 - p2) * m_direction >= 0)
					{
						m_intersection = p3;
						m_intersection_index = v3;
						m_new_u = m_scalar_field[v3];
						m_iedge1 = v3;
						m_iedge2 = v2;
					}
				else
					{
						m_intersection = p2;
						m_intersection_index = v2;
						m_new_u = m_scalar_field[v2];
						m_iedge1 = v2;
						m_iedge2 = v3;
					}
			}
		else if(other_v_pos == 1)
			{
				intersection_type = POINT;
				k3d::vector3 p3 = m_points[v3];
				k3d::vector3 p1 = m_points[v1];
				if((p1 - p3) * m_direction >= 0)
					{
						m_intersection = p1;
						m_intersection_index = v1;
						m_new_u = m_scalar_field[v1];
						m_iedge1 = v1;
						m_iedge2 = v3;
					}
				else
					{
						m_intersection = p3;
						m_intersection_index = v3;
						m_new_u = m_scalar_field[v3];
						m_iedge1 = v3;
						m_iedge2 = v1;
					}
			}
		else
			{
				intersection_type = POINT;
				k3d::vector3 p1 = m_points[v1];
				k3d::vector3 p2 = m_points[v2];
				if((p2 - p1) * m_direction >= 0)
					{
						m_intersection = p2;
						m_intersection_index = v2;
						m_new_u = m_scalar_field[v2];
						m_iedge1 = v2;
						m_iedge2 = v1;
					}
				else
					{
						m_intersection = p1;
						m_intersection_index = v1;
						m_new_u = m_scalar_field[v1];
						m_iedge1 = v1;
						m_iedge2 = v2;
					}
			}
	}

	// From triangle barycenter, find which edge or point intersects with m_direction
	unsigned long find_triangle_intersection(unsigned long v1, unsigned long v2, unsigned long v3, intersection_t& intersection_type)
	{
		unsigned long neighbouring_triangle = test_intersection(v1, v2, v3, 0, intersection_type);
		if(intersection_type != NONE)
			return neighbouring_triangle;

		neighbouring_triangle = test_intersection(v2, v3, v1, 1, intersection_type);
		if(intersection_type != NONE)
			return neighbouring_triangle;

		neighbouring_triangle = test_intersection(v3, v1, v2, 2, intersection_type);
		assert_warning(intersection_type != NONE);
		return neighbouring_triangle;
	}

	bool flow_distance_ok()
	{
		if(!m_test_flow_distance)
			return true;

		// Check all points against m_intersection
		double square_dist = m_gradient_distance * m_gradient_distance;
		for(unsigned long g = 0; g < m_gradient_flows.size(); ++g)
			{
				for(unsigned long i = 0; i < m_gradient_flows[g].size(); ++i)
					{
						const k3d::vector3 p = m_gradient_flows[g][i];
						// Comparing squares is faster
						if((m_intersection - p).Length2() < square_dist)
							// TODO : then check distance along mesh surface for better accuracy + do some cache to faster point comparison process
							return false;
					}
			}

		return true;
	}

	void gradient_flow_line(const index_t start_triangle, const k3d::vector3& start_point, const double start_u, points_t& point_list, indices_t& triangle_list, std::vector<double>& u_list)
	{
		// Down flow
		m_missed = false;
		m_current_triangle = start_triangle;
		m_intersection = start_point;
		m_new_u = start_u;

		m_direction = -m_orthogonal_vectors[m_current_triangle];
		intersection_t intersection_type;
		const unsigned long v1 = m_triangles[start_triangle][0];
		const unsigned long v2 = m_triangles[start_triangle][1];
		const unsigned long v3 = m_triangles[start_triangle][2];
		m_current_triangle = find_triangle_intersection(v1, v2, v3, intersection_type);

		do
			{
				m_direction = -m_orthogonal_vectors[m_current_triangle];
				if(intersection_type == EDGE)
					edge_intersection(intersection_type);
				else if(intersection_type == POINT)
					vertex_case(m_iedge1, m_iedge2, false, intersection_type);

				if(intersection_type != NONE)
					{
						point_list.push_back(m_intersection);
						triangle_list.push_back(m_current_triangle);
						u_list.push_back(m_new_u);
					}
			}
		while(intersection_type != NONE && m_intersection_index != m_flow_min && flow_distance_ok());

		std::reverse(point_list.begin(), point_list.end());
		std::reverse(triangle_list.begin(), triangle_list.end());
		std::reverse(u_list.begin(), u_list.end());

		// Up
		m_missed = false;
		m_current_triangle = start_triangle;
		m_intersection = start_point;
		m_new_u = start_u;

		m_gradient_flow_points[m_current_triangle].push_back(m_intersection);

		m_direction = m_orthogonal_vectors[m_current_triangle];
		m_current_triangle = find_triangle_intersection(v1, v2, v3, intersection_type);

		do
			{
				m_direction = m_orthogonal_vectors[m_current_triangle];
				if(intersection_type == EDGE)
					edge_intersection(intersection_type);
				else if(intersection_type == POINT)
					vertex_case(m_iedge1, m_iedge2, true, intersection_type);

				if(intersection_type != NONE)
					{
						point_list.push_back(m_intersection);
						triangle_list.push_back(m_current_triangle);
						u_list.push_back(m_new_u);
					}
			}
		while(intersection_type != NONE && m_intersection_index != m_flow_max && flow_distance_ok());
	}

	// Find which edge or point intersects with m_direction
	unsigned long iso_find_triangle_intersection(intersection_t& intersection_type)
	{
		const unsigned long v1 = m_triangles[m_current_triangle][0];
		const unsigned long v2 = m_triangles[m_current_triangle][1];
		const unsigned long v3 = m_triangles[m_current_triangle][2];
		const double u1 = m_scalar_field[v1];
		const double u2 = m_scalar_field[v2];
		const double u3 = m_scalar_field[v3];
		const k3d::vector3 p1 = m_points[v1];
		const k3d::vector3 p2 = m_points[v2];
		const k3d::vector3 p3 = m_points[v3];

		m_direction.Normalize();

		// Avoid current edge
		unsigned long index = 3;
		if(m_iedge2 == v1 && m_iedge1 == v2)
			index = 0;
		else if(m_iedge2 == v2 && m_iedge1 == v3)
			index = 1;
		else if(m_iedge2 == v3 && m_iedge1 == v1)
			index = 2;

		intersection_type = NONE;
		const double epsilon = 0.0001;
		if(index != 0)
			{
				if(u1 == m_new_u)
					{
						m_iedge1 = v1;
						m_iedge2 = v2;
						intersection_type = POINT;
						m_intersection = p1;
						return m_current_triangle;
					}

				if(u2 == m_new_u)
					{
						m_iedge1 = v1;
						m_iedge2 = v2;
						intersection_type = POINT;
						m_intersection = p2;
						return m_current_triangle;
					}

				if((u1 < m_new_u && m_new_u < u2) || (u2 < m_new_u && m_new_u < u1))
					{
						double coef = (m_new_u - u1) / (u2 - u1);
						k3d::vector3 candidate = p1 + (p2 - p1)*coef;
						k3d::vector3 cdir = candidate - m_intersection;

						// Normalize
						const double clen = cdir.Length();
						if(clen < epsilon)
							{
								// Point likely on edge
								intersection_type = NONE;
								return m_triangle_neighbours[m_current_triangle][0];
							}
						else
							{
								cdir /= clen;
								if(cdir * m_direction > 0.9)
									{
										intersection_type = EDGE;
										m_intersection = candidate;
										m_iedge1 = v1;
										m_iedge2 = v2;
										return m_triangle_neighbours[m_current_triangle][0];
									}
							}
					}
			}
		if(index != 1)
			{
				if(u2 == m_new_u)
					{
						m_iedge1 = v2;
						m_iedge2 = v3;
						intersection_type = POINT;
						m_intersection = p2;
						return m_current_triangle;
					}

				if(u3 == m_new_u)
					{
						m_iedge1 = v2;
						m_iedge2 = v3;
						intersection_type = POINT;
						m_intersection = p3;
						return m_current_triangle;
					}

				if((u2 < m_new_u && m_new_u < u3) || (u3 < m_new_u && m_new_u < u2))
					{
						double coef = (m_new_u - u2) / (u3 - u2);
						k3d::vector3 candidate = p2 + (p3 - p2) * coef;
						k3d::vector3 cdir = candidate - m_intersection;

						// Normalize
						const double clen = cdir.Length();
						if(clen < epsilon)
						{
							// Point likely on edge
							intersection_type = NONE;
							return m_triangle_neighbours[m_current_triangle][1];
						}
						else
						{
							cdir /= clen;
							if(cdir * m_direction > 0.9)
								{
									intersection_type = EDGE;
									m_intersection = candidate;
									m_iedge1 = v2;
									m_iedge2 = v3;
									return m_triangle_neighbours[m_current_triangle][1];
								}
						}
					}
			}
		if(index != 2)
			{
				if(u3 == m_new_u)
					{
						m_iedge1 = v3;
						m_iedge2 = v1;
						intersection_type = POINT;
						m_intersection = p3;
						return m_current_triangle;
					}

				if(u1 == m_new_u)
					{
						m_iedge1 = v3;
						m_iedge2 = v1;
						intersection_type = POINT;
						m_intersection = p3;
						return m_current_triangle;
					}

				if((u3 < m_new_u && m_new_u < u1) || (u1 < m_new_u && m_new_u < u3))
					{
						double coef = (m_new_u - u3) / (u1 - u3);
						k3d::vector3 candidate = p3 + (p1 - p3) * coef;
						k3d::vector3 cdir = candidate - m_intersection;

						// Normalize
						const double clen = cdir.Length();
						if(clen < epsilon)
						{
							// Point likely on edge
							intersection_type = NONE;
							return m_triangle_neighbours[m_current_triangle][2];
						}
						else
						{
							cdir /= clen;
							if(cdir * m_direction > 0.9)
								{
									intersection_type = EDGE;
									m_intersection = candidate;
									m_iedge1 = v3;
									m_iedge2 = v1;
									return m_triangle_neighbours[m_current_triangle][2];
								}
						}
					}
			}

		assert_not_reached();
		return 0;
	}

	// Turn around v1 to find which triangle is crossed by the flow
	void iso_vertex_case(unsigned long v1, unsigned long v_start, intersection_t& intersection_type)
	{
		unsigned long v_start2 = v_start;
		unsigned long v2 = v_start;

		m_direction = m_tangent_vectors[m_current_triangle];
		m_iedge1 = 0;
		m_iedge2 = 0;

		do
			{
				index_t p1 = m_triangles[m_current_triangle][0];
				index_t p2 = m_triangles[m_current_triangle][1];
				index_t p3 = m_triangles[m_current_triangle][2];
				double u1 = m_scalar_field[p1];
				double u2 = m_scalar_field[p2];
				double u3 = m_scalar_field[p3];

				intersection_type = NONE;
				if(v1 == p1)
					{
						if(((u2 < m_new_u && m_new_u < u3) || (u2 < m_new_u && m_new_u < u3)))
							{
								k3d::vector3 candidate = p2 + (p3 - p2) * (m_new_u - u2) / (u3 - u2);
								if((candidate - m_intersection) * m_direction > 0)
									{
										intersection_type = EDGE;
										m_intersection = candidate;
										//m_iedge1 = v2;
										//m_iedge2 = v3;
										m_current_triangle = m_triangle_neighbours[m_current_triangle][1];
										return;
									}
							}
						else if(u2 == m_new_u || u3 == m_new_u)
							{
								return;
							}

						m_current_triangle = m_triangle_neighbours[m_current_triangle][0];
						if(v2 == p2)
							v_start2 = p3;

						v2 = p2;
					}
				else if(v1 == p2)
					{
						if(((u3 < m_new_u && m_new_u < u1) || (u3 < m_new_u && m_new_u < u1)))
							{
								k3d::vector3 candidate = p3 + (p1 - p3) * (m_new_u - u3) / (u1 - u3);
								if((candidate - m_intersection) * m_direction > 0)
									{
										intersection_type = EDGE;
										m_intersection = candidate;
										//m_iedge1 = v3;
										//m_iedge2 = v1;
										m_current_triangle = m_triangle_neighbours[m_current_triangle][2];
										return;
									}
							}
						else if(u3 == m_new_u || u1 == m_new_u)
							{
								return;
							}

						m_current_triangle = m_triangle_neighbours[m_current_triangle][1];
						if(v2 == p3)
							v_start2 = p1;

						v2 = p3;
					}
				else if(v1 == p3)
					{
						if(((u1 < m_new_u && m_new_u < u2) || (u1 < m_new_u && m_new_u < u2)))
							{
								k3d::vector3 candidate = p1 + (p2 - p1) * (m_new_u - u1) / (u2 - u1);
								if((candidate - m_intersection) * m_direction > 0)
									{
										intersection_type = EDGE;
										m_intersection = candidate;
										//m_iedge1 = v1;
										//m_iedge2 = v2;
										m_current_triangle = m_triangle_neighbours[m_current_triangle][0];
										return;
									}
							}
						else if(u1 == m_new_u || u2 == m_new_u)
							{
								return;
							}

						m_current_triangle = m_triangle_neighbours[m_current_triangle][2];
						if(v2 == p1)
							v_start2 = p2;

						v2 = p1;
					}
				else
					assert_not_reached();

			}
		while(v2 != v_start2);

		assert_not_reached();
	}

	void isoparametric_flow_line(points_t& point_list, indices_t& triangle_list)
	{
		const unsigned long start_triangle = m_current_triangle;
		point_list.push_back(m_intersection);
		triangle_list.push_back(m_current_triangle);

		intersection_t intersection_type;

		m_iedge1 = 0;
		m_iedge2 = 0;
		m_direction = m_tangent_vectors[m_current_triangle];
		m_current_triangle = iso_find_triangle_intersection(intersection_type);
		if(intersection_type == POINT)
			iso_vertex_case(m_iedge1, m_iedge2, intersection_type);
		else if(intersection_type == NONE)
			{
				m_iedge1 = 0;
				m_iedge2 = 0;
				m_direction = m_tangent_vectors[m_current_triangle];
				m_current_triangle = iso_find_triangle_intersection(intersection_type);
			}

		point_list.push_back(m_intersection);
		triangle_list.push_back(m_current_triangle);

		do
			{
				m_direction = m_tangent_vectors[m_current_triangle];
				if(intersection_type == EDGE)
					m_current_triangle = iso_find_triangle_intersection(intersection_type);
				else if(intersection_type == POINT)
					iso_vertex_case(m_iedge1, m_iedge2, intersection_type);
				else if(intersection_type == NONE)
					{
						m_iedge1 = 0;
						m_iedge2 = 0;
						m_direction = m_tangent_vectors[m_current_triangle];
						m_current_triangle = iso_find_triangle_intersection(intersection_type);
					}

				if(intersection_type != NONE)
					{
						point_list.push_back(m_intersection);
						triangle_list.push_back(m_current_triangle);
					}
			}
		while(intersection_type != NONE && m_current_triangle != start_triangle);
	}

	bool isoparametric_seed_point(const double distance, k3d::vector3& seed_point, index_t& seed_triangle, const bool forward)
	{
		const unsigned long start_triangle = m_current_triangle;
		k3d::vector3 previous_point = m_intersection;
		index_t previous_triangle = m_current_triangle;

		intersection_t intersection_type;
		double length = 0;

		m_iedge1 = 0;
		m_iedge2 = 0;
		m_direction = m_tangent_vectors[m_current_triangle];
		if(!forward)
			m_direction = -m_direction;
		m_current_triangle = iso_find_triangle_intersection(intersection_type);
		if(intersection_type == POINT)
			iso_vertex_case(m_iedge1, m_iedge2, intersection_type);
		else if(intersection_type == NONE)
			{
				m_iedge1 = 0;
				m_iedge2 = 0;
				m_direction = m_tangent_vectors[m_current_triangle];
				if(!forward)
					m_direction = -m_direction;
				m_current_triangle = iso_find_triangle_intersection(intersection_type);
			}

		length += (m_intersection - previous_point).Length();
		if(length >= distance)
			{
				seed_point = previous_point + (m_intersection - previous_point) * distance / length;
				seed_triangle = previous_triangle;
				return true;
			}

		previous_point = m_intersection;
		previous_triangle = m_current_triangle;

		do
			{
				m_direction = m_tangent_vectors[m_current_triangle];
				if(!forward)
					m_direction = -m_direction;
				if(intersection_type == EDGE)
					m_current_triangle = iso_find_triangle_intersection(intersection_type);
				else if(intersection_type == POINT)
					iso_vertex_case(m_iedge1, m_iedge2, intersection_type);
				else if(intersection_type == NONE)
					{
						m_iedge1 = 0;
						m_iedge2 = 0;
						m_direction = m_tangent_vectors[m_current_triangle];
						if(!forward)
							m_direction = -m_direction;
						m_current_triangle = iso_find_triangle_intersection(intersection_type);
					}

				if(intersection_type != NONE)
					{
						const double current_length = (m_intersection - previous_point).Length();
						if((current_length > 0) && (length + current_length >= distance))
							{
								seed_point = previous_point + (m_intersection - previous_point) * (distance - length) / current_length;
								seed_triangle = previous_triangle;
								return true;
							}
						else
							{
								previous_point = m_intersection;
								previous_triangle = m_current_triangle;
								length += current_length;
							}
					}
			}
		while(intersection_type != NONE && m_current_triangle != start_triangle);

		return false;
	}

	double flow_length(const points_t& point_list, const bool loop = false)
	{
		if(!point_list.size())
			return 0;

		points_t::const_iterator point = point_list.begin();
		points_t::const_iterator end = point_list.end();

		k3d::vector3 previous_point = *point;
		double length = 0;
		for(++point; point != end; ++point)
			{
				length += (*point - previous_point).Length();
				previous_point = *point;
			}

		if(loop)
			length += (*point_list.begin() - previous_point).Length();

		return length;
	}

	void flow_divisions(const points_t& flow_points, const unsigned long divisions, points_t& salient_points, indices_t& salient_indices)
	{
		const double length = flow_length(flow_points, true);
		const double step = length / static_cast<double>(divisions);

		unsigned long t = 0;
		double current_length = 0;
		double previous_length = 0;
		k3d::vector3 previous_point;
		k3d::vector3 current_point = flow_points[0];
		for(unsigned long n = 0; n < divisions; ++n)
			{
				const double threshold = step * static_cast<double>(n + 1);

				while(current_length < threshold)
					{
						++t;
						previous_point = current_point;
						current_point = flow_points[t];
						previous_length = current_length;
						current_length += (current_point - previous_point).Length();
					}

				const double edge_size = current_length - previous_length;
				k3d::vector3 point = previous_point;
				if(edge_size != 0)
					point += (current_point - previous_point) * (threshold - previous_length) / edge_size;

				salient_points.push_back(point);
				salient_indices.push_back(t - 1);
			}
	}

private:
	// Check for new input
	bool m_input_changed;

	// Internal variables
	unsigned long m_flow_min;
	unsigned long m_flow_max;
	points_t m_points;
	typedef boost::array<unsigned long, 3> triangle_t;
	typedef std::vector<triangle_t> triangle_list_t;
	triangle_list_t m_triangles;

	std::vector<double> m_scalar_field;

	triangle_list_t m_triangle_neighbours;
	points_t m_orthogonal_vectors;
	points_t m_tangent_vectors;

	// Temp variables
	k3d::vector3 m_intersection;
	unsigned long m_iedge1;
	unsigned long m_iedge2;
	k3d::vector3 m_direction;
	double m_new_u;
	unsigned long m_current_triangle;
	unsigned long m_intersection_index;
	bool m_missed;
	unsigned long m_missed_index;

	std::vector<points_t> m_gradient_flow_points;
	std::vector<points_t> m_gradient_flows;
	std::vector< std::vector<unsigned long> > m_gradient_flows_triangles;
	std::vector< std::vector<double> > m_gradient_flows_values;

	index_t m_middle_triangle;
points_t norms1;
points_t norms2;

	bool m_test_flow_distance;
	double m_gradient_distance;

std::vector<k3d::vector3> m_debug_vector_list;
	k3d_measurement_property(unsigned long, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::with_constraint) m_longitudinal_divisions;
	k3d_measurement_property(unsigned long, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::with_constraint) m_latitudinal_divisions;
	k3d_measurement_property(unsigned long, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::with_constraint) m_min_index;
	k3d_measurement_property(unsigned long, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::with_constraint) m_max_index;
	k3d_measurement_property(double, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::no_constraint) m_h1;
	k3d_measurement_property(double, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::no_constraint) m_h2;
	k3d_measurement_property(double, k3d::immutable_name, k3d::change_signal, k3d::with_undo, k3d::local_storage, k3d::no_constraint) m_curvature_sensibility;
};

/////////////////////////////////////////////////////////////////////////////
// quadrilateral_remeshing_factory

k3d::iplugin_factory& quadrilateral_remeshing_factory()
{
	return quadrilateral_remeshing_implementation::get_factory();
}

} // namespace libk3dmesh


